# -!- makefile -!- ###########################################################
#
#			MicroEmacs Jasspa Distribution 1999.
#
#  System        : MicroEmacs
#  Description	 : Make file for Windows using Microsoft MSVC v6.0 development kit.
#  Created       : Sat Jan 24 00:01:40 1998
#
#  Last Modified : <010802.1209>
#
#  Notes
#       Build from the command line using nmake. 
#
#	Run "nmake -f win32v6.mak"      for optimised build produces ./me32.exe
#	Run "nmake -f win32v6.mak men"  for console support          ./men32.exe
#	Run "nmake -f win32v6.mak meu"  for url support              ./meu32.exe
#	Run "nmake -f win32v6.mak menu" for console & url support    ./menu32.exe
#	Run "nmake -f win32v6.mak med"  for debug build produces     ./med32.exe
#
#	Run "nmake -f win32v6.mak clean"      to clean source directory
#	Run "nmake -f win32v6.mak spotless"   to clean source directory even more
#
#  History
#
##############################################################################
#
# Microsoft MSCV 6.0 install directory
TOOLSDIR      = c:\Program Files\Microsoft Visual Studio\vc98
#
# Installation Directory
INSTDIR	      = c:\emacs
INSTPROGFLAGS = 
#
# Local Definitions
RM            = - erase
RC            =	rc
CC            =	cl
LD            =	link
INSTALL       =	copy
CDEBUG        =	-nologo -G5 -W3 -GX -Z7 -YX -Yd -Od -MLd -D_DEBUG
COPTIMISE     =	-nologo -G5 -YX -GX -O2 -DNDEBUG=1
CPROFILE      =	-nologo -G5 -YX -GX
CDEFS         = -D_WIN32 -I. "-I$(TOOLSDIR)\include"
LDFLAGS       = /NOLOGO /INCREMENTAL:no /MACHINE:IX86 /PDB:NONE "/LIBPATH:$(TOOLSDIR)\lib"
LIBS          = user32.lib gdi32.lib winspool.lib comdlg32.lib advapi32.lib shell32.lib
#
# Rules
.SUFFIXES: .c .o .od .on .ou .onu .rc .res

.c.o:	
	$(CC) $(COPTIMISE) $(CDEFS) -c $< -Fo$*.o
.c.on:	
	$(CC) $(COPTIMISE) $(CDEFS) -D_WINCON -c $< -Fo$*.on
.c.ou:	
	$(CC) $(COPTIMISE) $(CDEFS) -D_URLSUPP -c $< -Fo$*.ou
.c.onu:	
	$(CC) $(COPTIMISE) $(CDEFS) -D_WINCON -D_URLSUPP -c $< -Fo$*.onu
.c.od:	
	$(CC) $(CDEBUG) $(CDEFS) -c $< -Fo$*.od

.rc.res:	
	$(RC) -v -i . -i "$(TOOLSDIR)\include" -i "$(TOOLSDIR)\mfc\include" -fo $@ $*.rc 
#
# Source files
STDHDR	= ebind.h edef.h eextrn.h efunc.h emain.h emode.h eprint.h \
	  esearch.h eskeys.h estruct.h eterm.h evar.h evers.h \
	  ebind.def efunc.def eprint.def evar.def etermcap.def emode.def eskeys.def
STDSRC	= abbrev.c basic.c bind.c buffer.c crypt.c dirlist.c display.c \
	  eval.c exec.c file.c fileio.c hilight.c history.c input.c \
	  isearch.c key.c line.c macro.c main.c narrow.c next.c osd.c \
	  print.c random.c regex.c region.c registry.c search.c spawn.c \
	  spell.c tag.c termio.c time.c undo.c window.c word.c

PLTHDR  = wintermr.h
PLTSRC  = winterm.c winprint.c
PLTRC   = me.rc
#
# Object files
STDOBJ	= $(STDSRC:.c=.o)	
PLTOBJ  = $(PLTSRC:.c=.o)
OBJ	= $(STDOBJ) $(PLTOBJ)

NSTDOBJ	= $(STDSRC:.c=.on)	
NPLTOBJ = $(PLTSRC:.c=.on)
NOBJ	= $(NSTDOBJ) $(NPLTOBJ)

USTDOBJ	= $(STDSRC:.c=.ou)	
UPLTOBJ = $(PLTSRC:.c=.ou)
UOBJ	= $(USTDOBJ) $(UPLTOBJ)

NUSTDOBJ = $(STDSRC:.c=.onu)	
NUPLTOBJ = $(PLTSRC:.c=.onu)
NUOBJ	= $(NUSTDOBJ) $(NUPLTOBJ)

DSTDOBJ	= $(STDSRC:.c=.od)	
DPLTOBJ = $(PLTSRC:.c=.od)
DOBJ	= $(DSTDOBJ) $(DPLTOBJ)

PLTRES  = $(PLTRC:.rc=.res)
#
# Targets
all: me

install: me
	$(INSTALL) $(INSTPROGFLAGS) me $(INSTDIR)
	@echo "install done"

clean:
	$(RM) me32.exe
	$(RM) med32.exe
	$(RM) men32.exe
	$(RM) meu32.exe
	$(RM) menu32.exe
	$(RM) *.o
	$(RM) *.od
	$(RM) *.on
	$(RM) *.ou 
	$(RM) *.onu
	$(RM) me.aps
	$(RM) me.map
	$(RM) $(PLTRES)

spotless: clean
	$(RM) *.opt
	$(RM) *.pch
	$(RM) tags
	$(RM) *~

me:	me32.exe
me32.exe: $(OBJ) $(PLTRES)
	$(LD) $(LDFLAGS) /SUBSYSTEM:windows /out:$@ $(OBJ) $(PLTRES) $(LIBS)
	
men:	men32.exe
men32.exe: $(NOBJ) $(PLTRES)
	$(LD) $(LDFLAGS) /SUBSYSTEM:console /out:$@ $(NOBJ) $(PLTRES) $(LIBS)
	
meu:	meu32.exe
meu32.exe: $(UOBJ) $(PLTRES)
	$(LD) $(LDFLAGS) /SUBSYSTEM:windows /out:$@ $(UOBJ) $(PLTRES) ws2_32.lib $(LIBS)
	
menu:	menu32.exe
menu32.exe: $(NUOBJ) $(PLTRES)
	$(LD) $(LDFLAGS) /SUBSYSTEM:console /out:$@ $(NUOBJ) $(PLTRES) ws2_32.lib $(LIBS)

med:	med32.exe
med32.exe: $(DOBJ) $(PLTRES)
	$(LD) $(LDFLAGS) /DEBUG /SUBSYSTEM:windows /out:$@ $(DOBJ) $(PLTRES) $(LIBS)
#
# Dependancies
$(STDOBJ): $(STDHDR)
$(PLTOBJ): $(STDHDR) $(PLTHDR)

$(NSTDOBJ): $(STDHDR)
$(NPLTOBJ): $(STDHDR) $(PLTHDR)

$(USTDOBJ): $(STDHDR)
$(UPLTOBJ): $(STDHDR) $(PLTHDR)

$(NUSTDOBJ): $(STDHDR)
$(NUPLTOBJ): $(STDHDR) $(PLTHDR)

$(DSTDOBJ): $(STDHDR)
$(DPLTOBJ): $(STDHDR) $(PLTHDR)

